// File: fig0731.cpp
// Computer Systems, Fourth Edition
// Figure 7.31
// A multiple-token parser

#include <iostream>
#include <cctype> // isalpha, isdigit
#include <string> // string, getline
using namespace std;

#include "inBuffer.hpp" // InBuffer
InBuffer inBuffer;

enum Token {eT_IDENTIFIER, eT_INTEGER, eT_EMPTY, eT_INVALID};

class AToken {
public:
   virtual Token tokenType () = 0;
   virtual void printToken () = 0;
};

class TEmpty : public AToken {
public:
   Token tokenType () { return eT_EMPTY; }
   void printToken () { cout << "Empty token" << endl; }
};

class TInvalid : public AToken {
public:
   Token tokenType () { return eT_INVALID; }
   void printToken () { cout << "Syntax error" << endl; }
};

class TInteger : public AToken {
private:
   int intValue;
public:
   TInteger (int i) { intValue = i; }
   Token tokenType () { return eT_INTEGER; }
   void printToken () { cout << "Integer    = " << intValue << endl; }
};

class TIdentifier : public AToken {
private:
   string identValue;
public:
   TIdentifier (string str) { identValue = str; }
   Token tokenType () { return eT_IDENTIFIER; }
   void printToken () { cout << "Identifier = " << identValue << endl; }
};

enum State {eS_START, eS_IDENT, eS_SIGN, eS_INTEGER, eS_STOP};

void getToken (AToken*& pAT) {
   // Pre: pAT is set to a token object.
   // Post: pAT is set to a token object that corresponds to
   // the next token in the input buffer.
   State state;
   char nextChar;
   int sign;
   int localIntValue;
   string localIdentValue;
   delete pAT;
   pAT = new TEmpty;
   state = eS_START;
   do {
      inBuffer.advanceInput (nextChar);
      switch (state) {
      case eS_START:
         if (isalpha(nextChar)) {
            localIdentValue = nextChar;
            state = eS_IDENT;
         }
         else if (nextChar == '-') {
            sign = -1;
            state = eS_SIGN;
         }
         else if (nextChar == '+') {
            sign = +1;
            state = eS_SIGN;
         }
         else if (isdigit(nextChar)) {
            localIntValue = nextChar - '0';
            sign = +1;
            state = eS_INTEGER;
         }
         else if (nextChar == '\n') {
            state = eS_STOP;
         }
         else if (nextChar != ' ') {
            delete pAT;
            pAT = new TInvalid;
         }
         break;
      case eS_IDENT:
         if (isalpha (nextChar) || isdigit (nextChar)) {
            localIdentValue.push_back (nextChar);
         }
         else {
            inBuffer.backUpInput ();
            delete pAT;
            pAT = new TIdentifier (localIdentValue);
            state = eS_STOP;
         }
         break;
      case eS_SIGN:
         if (isdigit (nextChar)) {
            localIntValue = nextChar - '0';
            state = eS_INTEGER;
         }
         else {
            delete pAT;
            pAT = new TInvalid;
         }
         break;
      case eS_INTEGER:
         if (isdigit (nextChar)) {
            localIntValue = 10 * localIntValue + nextChar - '0';
         }
         else {
            inBuffer.backUpInput ();
            delete pAT;
            pAT = new TInteger (sign * localIntValue);
            state = eS_STOP;
         }
         break;
      }
   }
   while ((state != eS_STOP) && (pAT->tokenType () != eT_INVALID));
}

int main () {
   AToken* pAToken = new TEmpty;
   inBuffer.getLine ();
   while (!cin.eof ()) {
      do {
         getToken (pAToken);
         pAToken->printToken ();
      }
      while ((pAToken->tokenType () != eT_EMPTY)
          && (pAToken->tokenType () != eT_INVALID));
      inBuffer.getLine ();
   }
   delete pAToken;
   return 0;
}
