// File: fig0735.cpp
// Computer Systems, Fourth Edition
// Figure 7.35
// A language translator

#include <iostream>
#include <cctype> // isalpha, isdigit
#include <string> // string, getline
#include <map>
using namespace std;

#include "inBuffer.hpp" // InBuffer
InBuffer inBuffer;

// ======== Global tables
enum Mnemon {
   eM_ADD, eM_SUB, eM_MUL, eM_DIV,
   eM_NEG, eM_SET, eM_END, eM_EMPTY
};

map <Mnemon, string> operatorTable;
map <string, Mnemon> mnemonTable;

void initGlobalTables () {
   operatorTable [eM_ADD] = "+";
   operatorTable [eM_SUB] = "-";
   operatorTable [eM_MUL] = "*";
   operatorTable [eM_DIV] = "/";
   mnemonTable ["add"] = eM_ADD;
   mnemonTable ["sub"] = eM_SUB;
   mnemonTable ["mul"] = eM_MUL;
   mnemonTable ["div"] = eM_DIV;
   mnemonTable ["neg"] = eM_NEG;
   mnemonTable ["set"] = eM_SET;
   mnemonTable ["end"] = eM_END;
}

void lookUpMnemon (string id, Mnemon& mn, bool& fnd) {
   string lowerId = "";
   for (int i = 0; i < id.size (); i++) {
      lowerId.push_back (tolower (id[i]));
   }
   if (mnemonTable.count (lowerId) == 1) {
      mn = mnemonTable[lowerId];
      fnd = true;
   }
   else {
      fnd = false;
   }
}

// ======== The argument classes
class AArg {
public:
   virtual void putArg () = 0;
};

class IdentArg : public AArg {
private:
   string identValue;
public:
   IdentArg (string str) { identValue = str; }
   void putArg () { cout << identValue; }
};

class IntArg : public AArg {
private:
   int intValue;
public:
   IntArg (int i) { intValue = i; }
   void putArg () { cout << intValue; }
};

// ======== The token classes
enum Token {
   eT_IDENTIFIER, eT_INTEGER, eT_COMMA, eT_LEFT_PAREN,
   eT_RIGHT_PAREN, eT_EMPTY, eT_INVALID
};

class AToken {
public:
   virtual Token tokenType () = 0;
   virtual AArg* getArg () { return 0; }
   virtual string getIdentValue () { return ""; }
};

class TIdentifier : public AToken {
private:
   string identValue;
public:
   TIdentifier (string str) { identValue = str; }
   AArg* getArg () { 
      AArg* pArg = new IdentArg (identValue);
      return pArg;
   }
   Token tokenType () { return eT_IDENTIFIER; }
   string getIdentValue () { return identValue; }
};

class TInteger : public AToken {
private:
   int intValue;
public:
   TInteger (int i) { intValue = i; }
   AArg* getArg () {
      IntArg* pArg = new IntArg (intValue);
      return pArg;
   }
   Token tokenType () { return eT_INTEGER; }
};

class TComma : public AToken {
public:
   Token tokenType () { return eT_COMMA; }
};

class TLeftParen : public AToken {
public:
   Token tokenType () { return eT_LEFT_PAREN; }
};

class TRightParen : public AToken {
public:
   Token tokenType () { return eT_RIGHT_PAREN; }
};

class TEmpty : public AToken {
public:
   Token tokenType () { return eT_EMPTY; }
};

class TInvalid : public AToken {
public:
   Token tokenType () { return eT_INVALID; }
};

// ======== The code classes
class ACode {
public:
   virtual bool isError () = 0;
   virtual void generateCode () = 0;
};

// ======== The error code classes
class Error : public ACode {
public:
   bool isError () { return true; }
};

class ENoEnd : public Error {
public:
   void generateCode () {
      cout << "Error: Missing \"end\" sentinel.";
   }
};

class ENoFunct : public Error {
public:
   void generateCode () {
      cout << "Error: Line must begin with function identifier.";
   }
};

class ENoLeft : public Error {
public:
   void generateCode () {
      cout << "Error: Left parenthesis expected after function.";
   }
};

class EBad1st : public Error {
public:
   void generateCode () {
      cout << "Error: First argument not an identifier.";
   }
};

class EBadNeg : public Error {
public:
   void generateCode () {
      cout << "Error: The argument of \"neg\" is malformed.";
   }
};

class ENoComma : public Error {
public:
   void generateCode () {
      cout << "Error: Comma expected after first argument.";
   }
};

class EBad2nd : public Error {
public:
   void generateCode () {
      cout << "Error: Second argument not an identifier or integer.";
   }
};

class ENoRight : public Error {
public:
   void generateCode () {
      cout << "Error: Right parenthesis expected after argument.";
   }
};

class EExtraChars : public Error {
public:
   void generateCode () {
      cout << "Error: Illegal trailing character.";
   }
};

// ======== The valid code classes
class Valid : public ACode {
protected:
   Mnemon mnemonic;
public:
   bool isError () { return false; }
};

class ZeroArg : public Valid {
public:
   ZeroArg (Mnemon mn) { mnemonic = mn; }
   void generateCode () { } // cout nothing
};

class OneArg : public Valid {
private:
   AArg *pFirstArg;
public:
   OneArg (Mnemon mn, AArg* pFArg) { 
      mnemonic = mn;
      pFirstArg = pFArg;
   }
   ~OneArg () { delete pFirstArg; }
   void generateCode () {
      pFirstArg->putArg ();
      cout << " := -";
      pFirstArg->putArg ();
   }
};

class TwoArgs : public Valid {
private:
   AArg *pFirstArg, *pSecondArg;
public:
   TwoArgs (Mnemon mn, AArg* pFArg, AArg* pSArg) { 
      mnemonic = mn;
      pFirstArg = pFArg;
      pSecondArg = pSArg;
   }
   ~TwoArgs () { 
      delete pFirstArg;
      delete pSecondArg;
   }
   void generateCode () {
      switch (mnemonic) {
      case eM_SET:
         pFirstArg->putArg ();
         cout << " := ";
         pSecondArg->putArg ();
         break;
      case eM_ADD: case eM_SUB: case eM_MUL: case eM_DIV:
         pFirstArg->putArg ();
         cout << " := ";
         pFirstArg->putArg ();
         cout << " " << operatorTable[mnemonic] << " ";
         pSecondArg->putArg ();
         break;
      }
   }
};

// ======== The lexical analyzer
enum LexState {
   eLS_START, eLS_IDENT, eLS_SIGN, eLS_INTEGER, eLS_STOP
};

void getToken (AToken*& pAT) {
   // Pre: pAT is allocated to an irrelevant value.
   char nextChar;
   string localIdentValue;
   int localIntValue;
   int sign;
   delete pAT;
   pAT = new TEmpty;
   LexState state = eLS_START;
   do {
      inBuffer.advanceInput (nextChar);
      switch (state) {
      case eLS_START:
         if (isalpha (nextChar)) {  
            localIdentValue = nextChar;
            state = eLS_IDENT;
         }
         else if (nextChar == '-') {
            sign = -1;
            state = eLS_SIGN;
         }
         else if (nextChar == '+') {
            sign = +1;
            state = eLS_SIGN;
         }
         else if (isdigit (nextChar)) {
            localIntValue = nextChar - '0';
            sign = +1;
            state = eLS_INTEGER;
         }
         else if (nextChar == ',') {
            delete pAT;
            pAT = new TComma;
            state = eLS_STOP;
         }
         else if (nextChar == '(') {
            delete pAT;
            pAT = new TLeftParen;
            state = eLS_STOP;
         }
         else if (nextChar == ')') {
            delete pAT;
            pAT = new TRightParen;
            state = eLS_STOP;
         }
         else if (nextChar == '\n') {
            // pAT is initialized to TEmpty.
            state = eLS_STOP;
         }
         else if (nextChar != ' ') {
            delete pAT;
            pAT = new TInvalid;
         }
         break;
      case eLS_IDENT:
         if (isalpha (nextChar) || isdigit (nextChar)) {
            localIdentValue.push_back (nextChar);
         }
         else {
            inBuffer.backUpInput();
            delete pAT;
            pAT = new TIdentifier (localIdentValue);
            state = eLS_STOP;
         }
         break;
      case eLS_SIGN:
         if (isdigit (nextChar)) {
            localIntValue = nextChar - '0';
            state = eLS_INTEGER;
         }
         else {
            delete pAT;
            pAT = new TInvalid;
         }
         break;
      case eLS_INTEGER:
         if (isdigit (nextChar)) {
            localIntValue = 10 * localIntValue + nextChar - '0';
         }
         else {
            inBuffer.backUpInput();
            delete pAT;
            pAT = new TInteger (sign * localIntValue);
            state = eLS_STOP;
         }
         break;
      }
   }
   while ((state != eLS_STOP) && (pAT->tokenType () != eT_INVALID));
}

// ======== The parser
enum ParseState {
   ePS_START, ePS_END, ePS_FUNCTION, ePS_OPEN, ePS_1ST_OPRND,
   ePS_UNARY, ePS_COMMA, ePS_2ND_OPRND, ePS_NON_UNARY, ePS_FINISH
};

void processSourceLine (ACode*& pAC, bool& term) {
   // Pre: term is false.
   // A source line is in inBuffer ready for processing.
   // pAC is allocated to an irrelevant value.
   AArg* pLocalFirstArg;
   AArg* pLocalSecondArg;
   Mnemon mnemon;
   delete pAC;
   pAC = new ZeroArg (eM_EMPTY);
   AToken* pAToken = new TEmpty;
   ParseState state = ePS_START;
   do {
      getToken (pAToken);
      switch (state) {
      case ePS_START:
         if (pAToken->tokenType () == eT_IDENTIFIER) {
            string tempStr = pAToken->getIdentValue ();
            bool found;
            lookUpMnemon (tempStr, mnemon, found);
            if (found) {
               if (mnemon == eM_END) {
                  delete pAC;
                  pAC = new ZeroArg (eM_END);
                  term = true;
                  state = ePS_END;
               }
               else {
                  state = ePS_FUNCTION;
               }
            }
            else {
               delete pAC;
               pAC = new ENoFunct;
            }
         }
         else if (pAToken->tokenType () == eT_EMPTY) {
            // pAC initialized to ZeroArg (eM_EMPTY)
            state = ePS_FINISH;
         }
         else {
            delete pAC;
            pAC = new ENoFunct;
         }
         break;
      case ePS_END:
         if (pAToken->tokenType () == eT_EMPTY) {
            state = ePS_FINISH;
         }
         else {
            delete pAC;
            pAC = new EExtraChars;
         }
         break;
      case ePS_FUNCTION:
         if (pAToken->tokenType () == eT_LEFT_PAREN) {
            state = ePS_OPEN;
         }
         else {
            delete pAC;
            pAC = new ENoLeft;
         }
         break;
      case ePS_OPEN:
         if (pAToken->tokenType () == eT_IDENTIFIER) {
            pLocalFirstArg = pAToken->getArg ();
            state = ePS_1ST_OPRND;
         }
         else {
            delete pAC;
            pAC = new EBad1st;
         }
         break;
      case ePS_1ST_OPRND:
         if (mnemon == eM_NEG) {
            if (pAToken->tokenType () == eT_RIGHT_PAREN) {
               delete pAC;
               pAC = new OneArg (mnemon, pLocalFirstArg);
               state = ePS_UNARY;
            }
            else {
               delete pAC;
               pAC = new EBadNeg;
            }
         }
         else if (pAToken->tokenType () == eT_COMMA) {
            state = ePS_COMMA;
         }
         else {
            delete pAC;
            pAC = new ENoComma;
         }
         break;
      case ePS_UNARY:
         if (pAToken->tokenType () == eT_EMPTY) {
            state = ePS_FINISH;
         }
         else {
            delete pAC;
            pAC = new EExtraChars;
         }
         break;
      case ePS_COMMA:
         if ((pAToken->tokenType () == eT_IDENTIFIER) 
         ||  (pAToken->tokenType () == eT_INTEGER)) {
            pLocalSecondArg = pAToken->getArg ();
            delete pAC;
            pAC = new TwoArgs (mnemon, pLocalFirstArg, pLocalSecondArg);
            state = ePS_2ND_OPRND;
         }
         else {
            delete pAC;
            pAC = new EBad2nd;
         }
         break;
      case ePS_2ND_OPRND:
         if (pAToken->tokenType () == eT_RIGHT_PAREN) {
            state = ePS_NON_UNARY;
         }
         else {
            delete pAC;
            pAC = new ENoRight;
         }
         break;
      case ePS_NON_UNARY:
         if (pAToken->tokenType () == eT_EMPTY) {
            state = ePS_FINISH;
         }
         else {
            delete pAC;
            pAC = new EExtraChars;
         }
         break;
      }
   }
   while ((state != ePS_FINISH) && (!pAC->isError ()));
   delete pAToken;
}

// ======== The main program
int main () {
   bool terminateWithEnd = false;
   ACode* pACode = new ZeroArg (eM_EMPTY);
   initGlobalTables ();
   inBuffer.getLine ();
   while (! (cin.eof() || terminateWithEnd)) {
      processSourceLine (pACode, terminateWithEnd);
      pACode->generateCode ();
      cout << endl;
      inBuffer.getLine ();
   }
   if (!terminateWithEnd) {
      delete pACode;
      pACode = new ENoEnd;
      pACode->generateCode ();
      cout << endl;
   }
   delete pACode;
   return 0;
}
